/***********************************************************
Copyright 1995 by Farzad Farid,
Paris, France.

                        All Rights Reserved

Permission to use, copy, modify, and distribute this software and its
documentation for any purpose and without fee is hereby granted,
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in
supporting documentation, and that the names of Farzad Farid not be
used in advertising or publicity pertaining to distribution of the
software without specific, written prior permission.

FARZAD FARID DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
EVENT SHALL FARZAD FARID BE LIABLE FOR ANY SPECIAL, INDIRECT OR
CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF
USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
PERFORMANCE OF THIS SOFTWARE.

******************************************************************/

/* Expect module */

/*
 * This module provides an interface to the Expect library by Don Libes
 *
 * $Log: expectmodule.c,v $
 * Revision 1.6  1995/08/20  17:47:34  farzy
 * Reformatted the source.
 *
 * Revision 1.5  1995/08/20  17:23:35  farzy
 * New match_max variable that sets the maximum number of chars to match.
 * Fixed Interect() where typing ^Z in the spawned process and then
 * returning would generate an "Interrupted system call" error.
 * New Close() method.
 * Fixed PyExpect_Expect() where a freed memory location was used.
 *
 * Revision 1.4  1995/08/15  16:42:56  farzy
 * Farzad Farid:
 * * New features:
 *   Started adding the code to make the module use the Python regexp
 * library. Does not work yet. Read the BETA file for more info.
 *
 * Saad Mufti:
 * * Bug fixes:
 *   Fixed bug in store_pattern that disallowed freeing memory on error.
 *
 *   Improved send method by looping until all data is written.
 *
 * * The expect method is now a module method instead of a method of an
 * expect object. Also, the name "expect object" should now more properly
 * be referred to as "spawn object". For details of the new expect method,
 * refer to the other docs.
 *
 * Revision 1.3  1995/04/08  05:09:47  farzy
 * Corrected bug caused by the signal handler. The module no longers sets
 * a signal handler to catch SIGCHLD, instead it uses waitpid() to test
 * if the spawned process is still alive.
 *
 * Revision 1.2.1.1  1995/04/07  22:44:52  farzy
 * Corrected the last functions that did not follow the new naming
 * conventions.
 * All functions and methods now have the 3rd field of the PyMethodDef
 * set to 1. As a consequence PyArg_Parse has been replaced by
 * PyArg_ParseTuple wherever possible.
 * Started using documentation strings.
 * New interact() method.
 *
 * Revision 1.2  1995/02/25  21:56:10  farzy
 * * Suppressed constants:
 *   expect.TIMEOUT
 *   expect.EOF
 *   expect.FULLBUFFER
 *
 * * New exceptions:
 *   expect.timeout    = 'timeout'
 *   expect.eof        = 'eof'
 *   expect.fullbuffer = 'full_buffer'
 *
 * * New expect pattern:
 *   expect.exp_null
 *
 * * Renamed global functions:
 *   expect.timeout    -> expect.setdeftimeout
 *   expect.loguser    -> expect.setdefloguser
 *   expect.fullbuffer -> expect.setdeffull_buffer
 *
 * * New global function:
 *   expect.setdefremove_nulls
 *
 * * New expect object members
 *   timeout       R/W : -1, 0 or positive integer
 *   full_buffer   R/W : 0 or 1
 *   remove_nulls  R/W : 0 or 1
 *   buffer        R/O : string
 *   match_index   R/O : integer
 *   match_len     R/O : integer
 *   match_end     R/O : integer
 *
 * Revision 1.1  1995/02/15  16:27:16  farzy
 * Initial revision
 *
 *
 */

#include <stdio.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/termio.h>  /* Used for setting the keyboard in raw mode */
#include <sys/ioctl.h>   /* Idem */
#include <unistd.h>
#include <time.h>
#include "Python.h"
#include "sysmodule.h"
#include "modsupport.h"
#include "structmember.h"
#include "myselect.h"
#include "regexpr.h"
#include "expect.h"
 

static char rcsid[] = "$Id: expectmodule.c,v 1.6 1995/08/20 17:47:34 farzy Exp $";


/* Increment used when creating the exp_case array */

#define EXP_INCREMENT 5

/* 
** MAGIC NUMBER, more than this many return objects not handled (for my coding
** convenience and laziness)
*/

#define MAX_NUM_RET_OBJS 200

/* Global variables holding the exception types for errors detected
   by this module (but not argument type or memory errors, etc.). */

static PyObject *PyExpect_Error,
                *PyExpect_Error_Timeout,
                *PyExpect_Error_Eof,
                *PyExpect_Error_FullBuffer;


/* Convenience function to raise an error according to errno
   and return a NULL pointer from a function. */

static PyObject *
PyExpect_Err()
{
	return PyErr_SetFromErrno(PyExpect_Error);
}


/* The Expect object */

typedef struct {
    PyObject_HEAD
    int exp_pid;		/* Spawned program pid */
    int exp_fd;		/* Spawned program file descriptor */
    int exp_alive;		/* Is the connection alive? */
    int exp_status;		/* Status of a closed connection */
    int exp_match_max;	/* The max no. of characters to match in this */
			/*		process			      */
    int exp_loguser;	/* Loguser value for this process */
    int exp_full_buffer;	/* Fullbuffer flag value for this process */
    int exp_remove_nulls;	/* Remove_nulls flag for this process */
    PyObject *exp_buffer;	/* Current buffer */
    int exp_match_index;	/* Index of the match in the buffer */
    int exp_match_len;	/* Length of the string match */
    int exp_match_end;
} PyExpectExpObject;


/* Default values for the loguser and full_buffer */

static int exp_def_timeout;
static int exp_def_loguser;
static int exp_def_full_buffer;
static int exp_def_remove_nulls;
static int exp_def_match_max;


/* A general purpose static buffer */
#define EXP_ARRAY_SIZE 2048
static char exp_array[EXP_ARRAY_SIZE];


/*
** structure type that hold parameters to the expect method
*/
struct ProcCase {
    PyExpectExpObject *ep;
    struct exp_case *ecases;
    int numcases, maxcases;	    
};


/* A forward reference to the Exptype type object.
   The Exptype variable contains pointers to various functions,
   some of which call newexpobject(), which uses Exptype, so
   there has to be a circular reference. */

staticforward PyTypeObject PyExpectExp_Type;


/* Macro to check that a Python object is an expect object */
#define PyExpect_Check(op) ((op)->ob_type == &PyExpectExp_Type)


/* This section is not ready yet */

#undef USE_PYTHON_REGEXP
#ifdef USE_PYTHON_REGEXP
/* The following variable and functions are wrappers for the regular
   expression functions used by libexpect. */

typedef struct {
    struct re_pattern_buffer re_patbuf;
    struct re_registers re_regs;
    char re_fastmap[256];
} PyExpectRegStruct;

char *exp_RegexpError;

char *Exp_RegexpCompile(pattern)
     char *pattern;
{
    PyExpectRegStruct *re;

#ifdef EXP_DEBUG
    fprintf(stderr, "\nExpectRegexpCompile: pattern '%s', ",
	    pattern);
#endif
    re = PyMem_NEW(PyExpectRegStruct, 1);
    if (!re) {
	exp_RegexpError = "Out of memory";
#ifdef EXP_DEBUG
	fprintf(stderr, "Ouf of memory\n");
#endif
	return NULL;
    }
    re->re_patbuf.buffer = NULL;
    re->re_patbuf.allocated = 0;
    re->re_patbuf.fastmap = re->re_fastmap;
    re->re_patbuf.translate = NULL;
    exp_RegexpError = re_compile_pattern(pattern,
					 strlen(pattern),
					 &re->re_patbuf);
    if (exp_RegexpError != NULL) {
	free(re);
#ifdef EXP_DEBUG
	fprintf(stderr, "compile failed\n");
#endif
	return NULL;
    }
#ifdef EXP_DEBUG
    fprintf(stderr, "buffer '%s', fastmap '%s'\n",
	    re->re_patbuf.buffer, re->re_patbuf.fastmap);
#endif
    return ((char*)re);
}

void Exp_RegexpFree(re)
     char *re;
{
#ifdef EXP_DEBUG
    fprintf(stderr, "\nExpectRegexpFree\n");
#endif
    if (((PyExpectRegStruct*)re)->re_patbuf.buffer != NULL)
	free(((PyExpectRegStruct*)re)->re_patbuf.buffer);
    free(re);
}

int Exp_RegexpMatch(re, buffer, buflen)
     char *re;
     char *buffer;
     int buflen;
{
    int result, range;

#ifdef EXP_DEBUG
    fprintf(stderr,
	    "\nExpectRegexpMatch: buffer '%s', buflen %d, regexp '%s', ",
	    buffer, buflen, ((PyExpectRegStruct*)re)->re_patbuf.buffer);
#endif
    /* We use re_search because the regexp is unanchored
       range is is the number of positions to try */
    range = buflen;
    result = re_search(&((PyExpectRegStruct*)re)->re_patbuf,
		       buffer, buflen, 0, range,
		       &((PyExpectRegStruct*)re)->re_regs);
    if (result < -1) {	/* Error */
	exp_RegexpError = "match failure";
#ifdef EXP_DEBUG
	fprintf(stderr, "match failure ");
#endif
    }
    if (result < 0) {	/* Failure or match not found */
#ifdef EXP_DEBUG
	fprintf(stderr, "not found\n");
#endif
	return 0;
    }
#ifdef EXP_DEBUG
    fprintf(stderr, "found\n");
#endif
    return 1;		/* Match found */
}

char *Exp_RegexpStartp(re, buffer, group)
     char *re;
     char *buffer;
     int group;
{
#ifdef EXP_DEBUG
    fprintf(stderr, "\nExpectRegexpStartp: group %d, start %d\n",
	    group, ((PyExpectRegStruct*)re)->re_regs.start[group]);
#endif
    return (buffer +
	    ((PyExpectRegStruct*)re)->re_regs.start[group]);
}

char *Exp_RegexpEndp(re, buffer, group)
     char *re;
     char *buffer;
     int group;
{
#ifdef EXP_DEBUG
    fprintf(stderr, "\nExpectRegexpEndp: group %d, end %d\n",
	    group, ((PyExpectRegStruct*)re)->re_regs.end[group]);
#endif
    return (buffer +
	    ((PyExpectRegStruct*)re)->re_regs.end[group]);
}
#endif /* USE_PYTHON_REGEXP */


/* Create a new expect object.
   This just creates the object and initializes it.
   If the creation fails, return NULL and set an exception (implicit
   in PyObject_NEW()). */

static PyExpectExpObject *
PyExpectExp_New(pid, fd)
     int pid;
     int fd;
{
    PyExpectExpObject *ep;
    ep = PyObject_NEW(PyExpectExpObject, &PyExpectExp_Type);
    if (ep != NULL) {
	ep->exp_pid          = pid;
	ep->exp_fd           = fd;
	ep->exp_alive        = 1;
	ep->exp_status       = 0;
	ep->exp_match_max    = exp_def_match_max;
	ep->exp_loguser      = exp_def_loguser;
	ep->exp_full_buffer  = exp_def_full_buffer;
	ep->exp_remove_nulls = exp_def_remove_nulls;
	ep->exp_buffer       = NULL;
	ep->exp_match_index  = 0;
	ep->exp_match_len    = 0;
	ep->exp_match_end    = 0;
    }
    return ep;
}


/*
** Do equality test on two expect (spawn) objects
*/
static int
PyExpectExp_compare(PyExpectExpObject *v, PyExpectExpObject *w)
{
    return (v->exp_pid - w->exp_pid);
}


/* Check if the spawned process is still alive,
   if yes: return 1
   if no: set the flag and store the status and return 0

   The first version of this module used signals and SIGCHLD to detect
   automatically the end of the spawned process, but this did interfere
   with other modules that could spawn a program */

static int PyExpect_Alive(ep)
     PyExpectExpObject *ep;
{
    int pid, status;

    if (!ep->exp_alive)
	return 0;
    pid = waitpid(ep->exp_pid, &status, WNOHANG);
    if (!pid)	/* It's alive */
	return 1;
    else if (pid == ep->exp_pid) { /* It died */
	ep->exp_alive = 0;
	ep->exp_status = status;
	return 0;
    } else		/* Should not happen... */
	return -1;
}


/* Read from the file descriptor
   Code stolen from fileobject.c but with a little bit more
   error checking :) */

static PyObject *
PyExpectExp_Read(ep, args)
     PyExpectExpObject *ep;
     PyObject *args;
{
    int n, n1, n2, n3;
    PyObject *v;
    int num_ready_fd ;
    fd_set select_fdset ;
    struct timeval select_timeout ;
	
    if (!PyExpect_Alive(ep)) {
	PyErr_SetString(PyExpect_Error, "Connection dead");
	return NULL;
    }
    if (PyArg_ParseTuple(args, ""))
	n = 0;
    else {
	PyErr_Clear();
	if (!PyArg_ParseTuple(args, "i", &n))
	    return NULL;
	if (n < 0) {
	    PyErr_SetString(PyExc_ValueError,
			    "Negative read count");
	    return NULL;
	}
    }

    n2 = n != 0 ? n : BUFSIZ;
    v = PyString_FromStringAndSize((char *)NULL, n2);
    if (v == NULL)
	return NULL;
    n1 = 0;

    Py_BEGIN_ALLOW_THREADS
    for (;;) {
	/* Use select for non-blocking reads */
	select_timeout.tv_sec  = 0;
	select_timeout.tv_usec = 0 ;
	FD_ZERO(&select_fdset);
	FD_SET(ep->exp_fd, &select_fdset);
	num_ready_fd = select(FD_SETSIZE,
			      &select_fdset, NULL, NULL,
			      &select_timeout) ;
	if (num_ready_fd == 0)
	    break;
	if (num_ready_fd < 0) {	/* select error? */
	    Py_DECREF(v);
	    return PyExpect_Err();
	}

	n3 = read(ep->exp_fd,
		  PyString_AS_STRING((PyStringObject *)v)+n1,
		  n2-n1);
	if (n3 == 0)
	    break;
	if (n3 < 0) {		/* Error check */
	    Py_DECREF(v);
	    return PyExpect_Err();
	}

	n1 += n3;
	if (n1 == n)
	    break;
	if (n == 0) {
	    n2 = n1 + BUFSIZ;
	    Py_BLOCK_THREADS
	    if (_PyString_Resize(&v, n2) < 0)
		return NULL;
	    Py_UNBLOCK_THREADS
	}
    }
    Py_END_ALLOW_THREADS
    if (n1 != n2)
	_PyString_Resize(&v, n1);

    return v;
}


/* Send a string to the program */

static PyObject *
PyExpectExp_Send(ep, args)
     PyExpectExpObject *ep;
     PyObject *args;
{
    char *str;
    int done = 0;
    int index = 0;
    int length, written;
    
    if (!PyExpect_Alive(ep)) {
	PyErr_SetString(PyExpect_Error, "Connection dead");
	return NULL;
    }
    if (!PyArg_ParseTuple(args, "s#;String expected", &str, &length))
	return NULL;
    
    while (!done) {
	written = write(ep->exp_fd, str + index, length - index);

	/*
	** error occurred
	*/
	if (written == -1)
	    return PyExpect_Err();

	index += written;

	done = (index >= length);
    }
    
    Py_INCREF(Py_None);
    return Py_None;
}


/*
** free all the memory in the procCases structure
*/

static void
freeProcCaseMem(procCases, numcases)
    struct ProcCase *procCases;
    int numcases;
{
    int index, index2;

    for (index = 0; index < numcases; index++) {
	Py_DECREF(procCases[index].ep);
	if (procCases[index].ecases) {
	    for (index2 = 0; index2 < procCases[index].numcases; index2++) {
		/*
		 ** Free the memory for compiled regular expressions
		 ** allocated by the expect library.
		 */
		if ((procCases[index].ecases[index2].type == exp_regexp) && 
		    (procCases[index].ecases[index2].re)) {
#ifdef USE_PYTHON_REGEXP
		    ExpectRegexpFree(procCases[index].ecases[index2].re);
#else
		    PyMem_DEL(procCases[index].ecases[index2].re);
#endif /* USE_PYTHON_REGEXP */
		}
	    }
	    PyMem_DEL(procCases[index].ecases);
	}
    }
    
    PyMem_DEL(procCases);
}


/*
** add the indicated cases to the list for the indicated proc
** return 0 on error after setting python exception
*/
static int
addEcasesToProc(struct ProcCase *procCases, int index, 
		struct exp_case *ecases, int numcases)
{
    int maxcases;
    struct exp_case *saveEcases;
    
    if ((numcases + procCases[index].numcases) > 
	procCases[index].maxcases) {
	maxcases = procCases[index].maxcases + 
	    (EXP_INCREMENT > numcases ? EXP_INCREMENT : numcases);
	
	if (procCases[index].ecases) {
	    /*
	     ** reserve one extra space for the end marker
	     */
	    saveEcases = procCases[index].ecases;
	    
	    PyMem_RESIZE(procCases[index].ecases, 
			 struct exp_case, maxcases + 1);
	    
	    if (!procCases[index].ecases) {
		procCases[index].ecases = saveEcases;
		PyErr_NoMemory();
		return 0;
	    }
	}
	else {
	    /*
	     ** reserve one extra space for the end marker
	     */
	    procCases[index].ecases = PyMem_NEW(struct exp_case, 
						maxcases + 1);
	    
	    if (!procCases[index].ecases) {
		PyErr_NoMemory();
		return 0;
	    }
	    procCases[index].ecases[0].type = exp_end;
	}
	procCases[index].maxcases = maxcases;
    }
    
    memcpy(procCases[index].ecases + procCases[index].numcases,
	   ecases, numcases * sizeof(struct exp_case));
    
    procCases[index].numcases += numcases;
    
    procCases[index].ecases[procCases[index].numcases].type = exp_end;
    
    return 1;
}


/*
** add cases for indicated expect object, return 0 on failure after setting
** Python exception
*/
static int addProcCases(procCases, maxproccases, numproccases, 
			expobject, ecases, numcases)
    struct ProcCase **procCases;
    int *maxproccases;
    int *numproccases;
    PyExpectExpObject *expobject;
    struct exp_case *ecases;
    int numcases;
{
    int index;
    struct ProcCase *saveVal = *procCases;
    
    for (index = 0; index < *numproccases; index++) {
	if (PyExpectExp_compare((*procCases)[index].ep, expobject) == 0) {
	    break;
	}
    }

    /*
     ** add new entry for proc
     */
    if (index == *numproccases) {
	if (*numproccases == *maxproccases) {
	    *maxproccases += EXP_INCREMENT;
	    PyMem_RESIZE(*procCases, struct ProcCase, *maxproccases);
	    if (*procCases == NULL) {
		*procCases = saveVal;
		PyErr_NoMemory();
		return 0;			/*** out of memory ***/
	    }
	    
	    for (index = *numproccases; index < *maxproccases; index++) {
		(*procCases)[index].ep = NULL;
		(*procCases)[index].ecases = NULL;
		(*procCases)[index].numcases = 0;
		(*procCases)[index].maxcases = 0;
	    }
	}
	
	(*procCases)[*numproccases].ep = expobject;
	Py_INCREF(expobject);
	
	index = *numproccases;
	(*numproccases)++;
    }
    
    /*
     ** add to cases for proc entry
     */
    if (!addEcasesToProc(*procCases, index, ecases, numcases)) {
	return 0;
    }

    return 1;
}


static int store_pattern(ep, maxsize, n, type, pattern, value)
     struct exp_case **ep;
     int *maxsize;
     int *n;
     int type;
     char *pattern;
     int value;
{
    struct exp_case *save_val = *ep;
    
    if ((*n) >= *maxsize) {
	*maxsize += EXP_INCREMENT;
	/* Always allocate one more for exp_end */
	PyMem_RESIZE(*ep, struct exp_case, *maxsize + 1);
	if (*ep == NULL) {/* Out of memory */
	    *ep = save_val;
	    return 0;
	}
    }
    (*ep)[*n].type    = type;
    (*ep)[*n].pattern = pattern;
    (*ep)[*n].value   = value;
    (*ep)[*n].re      = NULL;
    *n += 1;
    return 1;
}


static int addOneEcase(PyObject *caseobject, 
		       struct exp_case **ecases,
		       int *numcases,
		       int *maxcases,
		       PyObject **retObjArray, 
		       int *retObjArraySize)
{
    int type;
    char *pattern;
    PyObject *anobject, *stringlist;
    int j;
    
    /* Each tuple can be :
       (integer, string, object), or
       (integer, list of strings, object).
       We test for the fist case here... */
    if (PyArg_ParseTuple(caseobject, "isO", &type, &pattern, &anobject)) {
	if (*retObjArraySize == MAX_NUM_RET_OBJS) {
	    PyErr_SetString(PyExc_ValueError,
			    "Too many conditions to handle");
	    return 0;
	}
	retObjArray[*retObjArraySize] = anobject;
	(*retObjArraySize)++;
	
	if (!store_pattern(ecases, maxcases, numcases,
			   type, pattern, *retObjArraySize - 1)) {
	    PyErr_NoMemory();
	    return 0;
	}
	
	/* Go to next element of the tuple */
	return 1;
    }
    
    /* ... and we test for the second case here */
    PyErr_Clear() ;
    if (!PyArg_ParseTuple(caseobject, "iO!O", &type, &PyList_Type,
			  &stringlist, &anobject)) {
	PyErr_SetString(PyExc_TypeError,
			"Pattern must be (int, string, object) or (int, [string,...], object)");
	return 0;
	}

    /* Here we've got a list of strings, we store one
       expect case for each string */
    for (j = PyList_Size(stringlist) - 1 ; j >= 0 ; j--) {
	anobject = PyList_GetItem(stringlist, j);
	if (!PyString_Check(anobject)) {
	    PyErr_SetString(PyExc_TypeError,
			    "Pattern must be (int, string, object) or (int, [string,...], object)");
	    return 0;
	}
	if (*retObjArraySize == MAX_NUM_RET_OBJS) {
	    PyErr_SetString(PyExc_ValueError,
			    "Too many conditions to handle");
	    return 0;
	}
	retObjArray[*retObjArraySize] = anobject;
	(*retObjArraySize)++;
	if (!store_pattern(&ecases, &maxcases, &numcases, type,
			   PyString_AsString(anobject),
			   *retObjArraySize - 1)) {
	    
	    PyErr_NoMemory();
	    return 0;
	}
    }
    
    return 1;
}


/*
** form array of ecases, on error return NULL after setting Python exception
*/
static struct exp_case *
formEcases(PyObject *caseobject, int *numcases, PyObject **retObjArray, 
	   int *retObjArraySize)
{
    struct exp_case *ecases = NULL;
    int maxcases = 0;
    int index;
    int size;
    PyObject *listitem;

    if (PyTuple_Check(caseobject)) {
	/*
	** initial guess of size, could be more than one b/c the second element
	** of tuple could be list of strings
	*/
	maxcases = EXP_INCREMENT;
	ecases = PyMem_NEW(struct exp_case, maxcases);

	if (ecases == NULL) {
	    PyErr_NoMemory();
	    return NULL;			/*** out of memory ***/
	}

	if (!addOneEcase(caseobject, &ecases, numcases,
			 &maxcases, retObjArray, 
			 retObjArraySize)) {
	    if (ecases)
		PyMem_DEL(ecases);
	    return NULL;		/*** EXIT ***/
	}

	return ecases;
    }

    if (!PyList_Check(caseobject)) {
	PyErr_SetString(PyExc_TypeError, "invalid args");
	return NULL;		/*** EXIT ***/
    }

    if (!(size = PyList_Size(caseobject))) {
	PyErr_SetString(PyExc_ValueError, "list must be non-empty");
	return NULL;                /*** EXIT ***/
    }


    /*
    ** the extra munge factor is b/c some of the second tuple elements could be
    ** a list of strings
    */
    maxcases = size * 3 / 2;
    ecases = PyMem_NEW(struct exp_case, maxcases);
    if (ecases == NULL) {
	PyErr_NoMemory();
	return NULL;                    /*** out of memory ***/
    }

    for (index = 0; index < size; index++) {
	listitem = PyList_GetItem(caseobject, index);

	if (!PyTuple_Check(listitem)) {
	    PyErr_SetString(PyExc_TypeError, "invalid args");
	    if (ecases)
		PyMem_DEL(ecases);
	    return NULL;                /*** EXIT ***/
	}

	if (!addOneEcase(listitem, &ecases, numcases, &maxcases, retObjArray, 
			 retObjArraySize)) {
	    if (ecases)
		PyMem_DEL(ecases);
	    return NULL;		/*** EXIT ***/
	}
    }

    return ecases;
}


/* Put the keyboard in raw or cooked mode,
   0 = cooked, 1 = raw */
/* XXX : Not very portable.... yet */

static void set_raw(fd, t)
     int fd;
     int t;
{
    static struct termio tty ;
    struct termio tmp ;
    
    if (t) {
	ioctl(fd, TCGETA, &tty) ;
	tmp = tty ;
	tmp.c_lflag &= ~(ICANON|ECHO) ;
	tmp.c_cc[VMIN] = 1 ;
	tmp.c_cc[VTIME] = 0 ;
	ioctl(fd, TCSETA, &tmp) ;
    }
    else 
	ioctl(fd, TCSETA, &tty) ;
}


/* Interact with the user */
/* XXX : This is a first try and will no hold the comparison against
   Expect's Interact function :) */

static PyObject *
PyExpectExp_Interact(ep, args)
     PyExpectExpObject *ep;
     PyObject *args;
{
    char escape_char, c;
    int infd; /* file descriptor corresponding to sys.stdin */
    FILE *instream, *outstream;
    PyObject *infileobject, *outfileobject;
    PyObject *string;
    int num_ready_fd;
    fd_set select_readfdset;
    int i, n;
    
    if (!PyExpect_Alive(ep)) {
	PyErr_SetString(PyExpect_Error, "Connection dead");
	return NULL;
    }

    /* Set the escape character, NUL means no escape char */
    if (PyArg_ParseTuple(args, ""))
	escape_char = 0x1B; /* ESC */
    else {
	PyErr_Clear();
	if (!PyArg_ParseTuple(args, "c", &escape_char))
	    return NULL;
    }

    /* Get the standard input and output file descriptors and
       check that stdin is a tty before going on */
    infileobject = PySys_GetObject("stdin");
    if (!infileobject) {
	PyErr_SetString(PyExpect_Error,
			"Lost sys.stdin");
	return NULL;
    }
    /* This will fail if sys.stdin is not a file */
    instream = PyFile_AsFile(infileobject);
    if (!instream || !isatty(infd = fileno(instream))) {
	PyErr_SetString(PyExpect_Error,
			"Not connected to a terminal");
	return NULL;
    }
    outfileobject = PySys_GetObject("stdout");
    if (!outfileobject) {
	PyErr_SetString(PyExpect_Error,
			"Lost sys.stdout");
	return NULL;
    }
    outstream = PyFile_AsFile(outfileobject);
    /* Now infd should be the input tty file descriptor and
       outfileobject an object with a write method.
       outstream may be NULL if outfileobject is not a stream */
    
    /* Go into raw mode */
    set_raw(infd, 1);
    
    /* Now loop */
    while(1) {
	/* Block until there is something to read */
	FD_ZERO(&select_readfdset);
	FD_SET(ep->exp_fd, &select_readfdset);
	FD_SET(infd,       &select_readfdset);
	num_ready_fd = select(FD_SETSIZE,
			      &select_readfdset,
			      NULL, NULL,
			      NULL) ;
	if (!PyExpect_Alive(ep)) {
	    /* Don't forget to return to cooked mode */
	    set_raw(infd, 0);
	    PyErr_SetString(PyExpect_Error_Eof,
			    "EOF reached");
	    return NULL;
	}
	
	if (num_ready_fd == 0)
	    continue; /* Should not happen.. I hope */
	if (num_ready_fd < 0) {	/* select error? */
		if (errno != EINTR) {
			set_raw(infd, 0);
			return PyExpect_Err();
		}
		/*
		** interrupted system call
		*/
		else {
			continue;
		}
	}
	
	/* Read from the spawned program and
	   write to the standard output */
	if (FD_ISSET(ep->exp_fd, &select_readfdset)) {
	    n = read(ep->exp_fd, exp_array,
		     EXP_ARRAY_SIZE);
	    string = PyString_FromStringAndSize(
						exp_array, n);
	    i = PyFile_WriteObject(string, outfileobject,
				   Py_PRINT_RAW);
	    /* Flush only if it's a file stream */
	    if (outstream)
		fflush(outstream);
	    Py_DECREF(string);
	    if (i) {
		set_raw(infd, 0);
		PyErr_SetString(PyExpect_Error,
				"Write to sys.stdout failed");
		return NULL;
	    }
	}
	
	/* Read from the tty and
	   write to the spawned program */
	if (FD_ISSET(infd, &select_readfdset)) {
	    n = read(infd, &c, 1);
	    if (escape_char && (c == escape_char))
		break;
	    write(ep->exp_fd, &c, 1);
	}
    }
    
    /* Return to cooked mode */
    set_raw(infd, 0);
    
    Py_INCREF(Py_None);
    return Py_None;
}


/* Close a spawned program by closing the file descriptor and
   waiting for the process to terminate (this releases the 
   process slot also) */

static void CloseExpectObject(ep)
	PyExpectExpObject *ep;
{
	if (PyExpect_Alive(ep) == 1) {
	    close(ep->exp_fd);
		/* Now wait */
	/* turned off wait to match semantics of expect tool,
	   which doesn't do a wait.
		while (PyExpect_Alive(ep) == 1)
		;
	*/
	}

	PyExpect_Alive(ep);
}


static PyObject *
PyExpectExp_Close(ep)
     PyExpectExpObject *ep;
{
	CloseExpectObject(ep);

    Py_INCREF(Py_None);
    return Py_None;
}


/* List of methods for expect objects */

static PyMethodDef PyExpectExp_methods[] = {
	{ "read",	(PyCFunction)PyExpectExp_Read,     1 },
	{ "send",	(PyCFunction)PyExpectExp_Send,     1 },
	{ "interact",	(PyCFunction)PyExpectExp_Interact, 1 },
	{ "close", (PyCFunction)PyExpectExp_Close, 1 },
	{ NULL,		NULL }		/* sentinel */
};

/* List of member for expect objects */

#define OFF(x) offsetof(PyExpectExpObject, x)

static struct memberlist PyExpectExp_members[] = {
	{ "pid",	T_INT,		OFF(exp_pid),		RO },
	{ "fd",		T_INT,		OFF(exp_fd),		RO },
	{ "alive",	T_INT,		OFF(exp_alive),		RO },
	{ "status",	T_INT,		OFF(exp_status),	RO },
	{ "match_max",	T_INT,		OFF(exp_match_max)},
	{ "loguser",	T_INT,		OFF(exp_loguser)},
	{ "full_buffer",T_INT,		OFF(exp_full_buffer)},
	{ "remove_nulls", T_INT,	OFF(exp_remove_nulls)},
	{ "buffer",	T_OBJECT,	OFF(exp_buffer),	RO },
	{ "match_index", T_INT,		OFF(exp_match_index),	RO },
	{ "match_len",	T_INT,		OFF(exp_match_len),	RO },
	{ "match_end",	T_INT,		OFF(exp_match_end),	RO },
	{ NULL }
};



/* Deallocate an expect object in response to the last Py_DECREF().
   First close the file descriptor. */

static void
PyExpectExp_dealloc(PyExpectExpObject *ep)
{
    CloseExpectObject(ep);
    PyMem_DEL(ep);
}


/* Return an expect object's named attribute. */

static PyObject *
PyExpectExp_getattr(PyExpectExpObject *ep, char *name)
{
    PyObject *res;

    res = Py_FindMethod(PyExpectExp_methods, (PyObject *) ep, name);
    if (res != NULL)
	return res;
    PyErr_Clear();
    /* XXX Hack: update manually the 'alive' and 'status' members */
    PyExpect_Alive(ep);
    return PyMember_Get((char *) ep, PyExpectExp_members, name);
}


/* Change an expect object's named attribute.
   Code based on structmember.c:setmember() */

static int
PyExpectExp_setattr(PyExpectExpObject *ep, char *name,
		    PyObject *value)
{
    struct memberlist *l;
    int newvalue;
	
    /* Get rid of read-only and non existent members to concentrate
       only on the members that can be modified */
    for (l = PyExpectExp_members; l->name != NULL; l++) {
	/* Loop again if it's not the right member */
	if (strcmp(l->name, name))
	    continue;
	
	if (l->readonly || l->type == T_STRING) {
	    PyErr_SetString(PyExc_TypeError,
			    "readonly attribute");
	    return -1;
	}
	if (value == NULL && l->type != T_OBJECT) {
	    PyErr_SetString(PyExc_TypeError,
			    "can't delete numeric/char attribute");
	    return -1;
	}
	
	/* We've got a changeable member, check the
	   new value before setting it */
	/* First sanity check: must be an integer */
	if (!PyArg_Parse(value, "i", &newvalue))
	    return -1;
	/* Now which member is it? */
	if (strcmp(name, "match_max") == 0) {
	    if (newvalue > 0) {
		ep->exp_match_max = newvalue;
	    } else {
		PyErr_SetString(PyExc_ValueError,
				"match_max value must be greater than 0");
		return -1;
	    }
	} else if (strcmp(name, "loguser") == 0) {
	    if (newvalue == 0 || newvalue == 1)
		ep->exp_loguser = newvalue;
	    else {
		PyErr_SetString(PyExc_ValueError,
				"Loguser value must be 0 or 1");
		return -1;
	    }
	} else if (strcmp(name, "full_buffer") == 0) {
	    if (newvalue == 0 || newvalue == 1)
		ep->exp_full_buffer = newvalue;
	    else {
		PyErr_SetString(PyExc_ValueError,
				"Full_buffer value must be 0 or 1");
		return -1;
	    }
	} else if (strcmp(name, "remove_nulls") == 0) {
	    if (newvalue == 0 || newvalue == 1)
		ep->exp_remove_nulls = newvalue;
	    else {
		PyErr_SetString(PyExc_ValueError,
				"Remove_nulls value must be 0 or 1");
		return -1;
	    }
	} else {
	    /* ??? If we get there we've
	       got an internal error */
	    PyErr_SetString(PyExc_AttributeError, name);
	    return -1;
	}
	return 0;
    }
    
    PyErr_SetString(PyExc_AttributeError, name);
    return -1;
}

/* Type object for expect objects. */

static PyTypeObject PyExpectExp_Type = {
	PyObject_HEAD_INIT(&PyType_Type)
	0,				/*ob_size*/
	"expect",			/*tp_name*/
	sizeof(PyExpectExpObject), 	/*tp_basicsize*/
	0,				/*tp_itemsize*/
	(destructor)PyExpectExp_dealloc, /*tp_dealloc*/
	0,				/*tp_print*/
	(getattrfunc)PyExpectExp_getattr, /*tp_getattr*/
	(setattrfunc)PyExpectExp_setattr, /*tp_setattr*/
	(cmpfunc)PyExpectExp_compare,	/*tp_compare*/
	0,				/*tp_repr*/
	0,				/*tp_as_number*/
	0,				/*tp_as_sequence*/
	0,				/*tp_as_mapping*/
	0,				/*tp_hash*/
};


/* Spawn a new process
   Return a new expect object. */

static PyObject *
PyExpect_Spawn(PyObject *self, PyObject *args)
{
    PyExpectExpObject *ep;
    PyObject *arguments;
    char *progname;
    char **argvlist;
    int fd, argc, i;
    PyObject *(*Py_GetItem) Py_PROTO((PyObject *, int));
    
    arguments = NULL;
    if (!PyArg_ParseTuple(args, "sO", &progname, &arguments))
	return NULL;
    if (PyTuple_Check(arguments)) {
	Py_GetItem = PyTuple_GetItem;
	argc = PyTuple_Size(arguments);
    } else if (PyList_Check(arguments)) {
	Py_GetItem = PyList_GetItem;
	argc = PyList_Size(arguments);
    } else {
	PyErr_SetString(PyExc_TypeError,
			"Argument must be tuple or list");
	return NULL;
    }
    
    /* Prepare the arguments array for exp_spawnv */
    argvlist = PyMem_NEW(char *, argc + 1);
    if (!argvlist)
	return PyErr_NoMemory();
    for (i = 0 ; i < argc ; i++) {
	if (!PyArg_Parse(Py_GetItem(arguments, i),
			 "s", &argvlist[i])) {
	    PyMem_DEL(argvlist);
	    PyErr_BadArgument();
	    return NULL;
	}
    }
    argvlist[argc] = NULL;
    
    fd = exp_spawnv(progname, argvlist);
    PyMem_DEL(argvlist);
    if (fd < 0)
	return PyExpect_Err();
    
    /* Create and return an expect object */
    ep = PyExpectExp_New(exp_pid, fd);
    /* If the object can't be created, don't forget to close the
       file descriptor again! */
    if (ep == NULL) {
	close(fd);
	waitpid(exp_pid, NULL, 0);
    }
    return (PyObject *) ep;
}


/* Wait for the output of a process to match a pattern */

static PyObject *
PyExpect_Expect(self, args)
     PyObject *self;
     PyObject *args;
{
    PyObject *retSpawnObj;
    PyObject *caselist, *caseobject;
    PyObject *object1, *object2, *expobject, *explist;
    PyObject *returnObjArray[MAX_NUM_RET_OBJS];
    /* array of return objects */
    int returnObjArraySize = 0;
    /* number of item in returnObjArray */
    struct ProcCase *procCases;
    int numproccases, maxproccases, numprocs;
    int index, index2, n, result;
    long timeout;
    struct exp_case *ecases = NULL;
    int numcases = 0;
    time_t curTime, newTime;
    fd_set spawnFdSet, spawnReadyFdSet;
    int maxFd = 0;
    int numready;		/* number of descriptors ready for reading */
    struct timeval selectTimeOut;
    int old_exp_timeout = exp_timeout;
    /*
    ** remember old value so we can restore it before
    ** returning
    */

    procCases = NULL;

    curTime = time(NULL);

    if (curTime == -1) {
	return PyExpect_Err();
    }

    if (!PyArg_ParseTuple(args, "O!l", &PyList_Type, &caselist, &timeout)) {
	PyErr_Clear();

	if (!PyArg_ParseTuple(args, "O!", &PyList_Type, &caselist))
	    return NULL;			/*** EXIT ***/
	
	timeout = exp_def_timeout;
    }

    if (timeout < -1) {
	PyErr_SetString(PyExc_ValueError, "timeout value must be >= -1");
	return NULL;
    }

    if (!(n = PyList_Size(caselist))) {
	PyErr_SetString(PyExc_ValueError, "Null length list");
	return NULL;			    /*** EXIT ***/
    }

    numproccases = 0;
    maxproccases = n;	    /*** initial guess ***/
    procCases = PyMem_NEW(struct ProcCase, maxproccases);
    if (procCases == NULL)
	return PyErr_NoMemory();

    for (index = 0; index < maxproccases; index++) {
	procCases[index].ep = NULL;
	procCases[index].ecases = NULL;
	procCases[index].numcases = 0;
	procCases[index].maxcases = 0;
    }

    for (index = 0; index < n; index++) {
	caseobject = PyList_GetItem(caselist, index);
	if (!PyTuple_Check(caseobject)) {
	    PyErr_SetString(PyExc_TypeError,
			    "Pattern must be size 2 tuple");
	    freeProcCaseMem(procCases, numproccases);
	    return NULL;
	}
	
	/* Each tuple must be :
	   (list of expect objects(or one expect object),
	   list of case tuples(or one case tuple).
	   We test for this case here... */
	
	if (!PyArg_ParseTuple(caseobject, "OO", &object1, &object2)) {
	    PyErr_SetString(PyExc_TypeError, "invalid arguments");
	    freeProcCaseMem(procCases, numproccases);
	    return NULL;
	}
	
	if (ecases)
	    PyMem_DEL(ecases);
	
	numcases = 0;
	if (!(ecases = formEcases(object2, &numcases, returnObjArray, 
				  &returnObjArraySize))) {
	    freeProcCaseMem(procCases, numproccases);
	    return NULL;
	}
	
	/*
	 ** add if a single expect object
	 */
	if (PyExpect_Check(object1)) {
	    if (!PyExpect_Alive(object1)) {
		PyErr_SetString(PyExpect_Error, "Connection dead");
		PyMem_DEL(ecases);
		freeProcCaseMem(procCases, numproccases);
		return NULL;
	    }
	    
	    if (!addProcCases(&procCases, &maxproccases, 
			      &numproccases, object1, ecases,
			      numcases)) {
		PyMem_DEL(ecases);
		freeProcCaseMem(procCases, numproccases);
		return NULL;
	    }
	}
	/*
	** it is a list of expect objects, iterate thru them adding
	** cases for each
	*/
	else if (PyList_Check(object1)) {
	    explist = object1; 
	    if (!(numprocs = PyList_Size(explist))) {
		PyErr_SetString(PyExc_ValueError, "Null length list");
		PyMem_DEL(ecases);
		freeProcCaseMem(procCases, numproccases);
		return NULL;			    /*** EXIT ***/
	    }
	    for (index2 = 0; index2 < numprocs; index2++) {
		expobject = PyList_GetItem(explist, index2);
		
		if (!PyExpect_Check(expobject)) {
		    PyErr_SetString(PyExc_TypeError, 
				    "invalid arguments");
		    PyMem_DEL(ecases);
		    freeProcCaseMem(procCases, numproccases);
		    return NULL;
		}
		
		if (!PyExpect_Alive(expobject)) {
		    PyErr_SetString(PyExpect_Error, 
				    "Connection dead");
		    PyMem_DEL(ecases);
		    freeProcCaseMem(procCases, numproccases);
		    return NULL;
		}
		
		
		if (!addProcCases(&procCases, &maxproccases, &numproccases, 
				  expobject, ecases, numcases)) {
		    PyMem_DEL(ecases);
		    freeProcCaseMem(procCases, numproccases);
		    return NULL;
		}
	    }
	}
	/*
	** error, object is of unknown type
	*/
	else {
	    PyErr_SetString(PyExc_TypeError, "invalid arguments");
	    PyMem_DEL(ecases);
	    freeProcCaseMem(procCases, numproccases);
	    return NULL;
	}
    }
    
    if (ecases)
	PyMem_DEL(ecases);
    
    numcases = 0;
    
    
    /*
    ** Form list of file descriptors id's for select
    */
    FD_ZERO(&spawnFdSet);
    for (index = 0; index < numproccases; index++) {
	FD_SET(procCases[index].ep->exp_fd, &spawnFdSet);
	if (procCases[index].ep->exp_fd > maxFd)
	    maxFd = procCases[index].ep->exp_fd;
    }
	

    /*
    ** adjust remaining time for time taken so far
    */
    if (timeout != -1) {
	newTime = time(NULL);
	
	if (newTime == -1) {
	    freeProcCaseMem(procCases, numproccases);
	    return PyExpect_Err();
	}
	
	timeout -= (newTime - curTime);
	
	if (timeout < 0)
	    timeout = 0;
	
	curTime = newTime;
    }


    /*
    ** why doesn't 0 work here
    */
    exp_timeout = 0;

    /*
    ** loop while time is left and a match hasn't been made
    */
    do {
	selectTimeOut.tv_sec = timeout;
	selectTimeOut.tv_usec = 0;
	spawnReadyFdSet = spawnFdSet;
	
	numready = select(maxFd + 1, &spawnReadyFdSet, NULL, NULL, 
			  timeout == -1 ? NULL : &selectTimeOut);
	
	if (numready == -1 && errno != EINTR) {
	    freeProcCaseMem(procCases, numproccases);
	    exp_timeout = old_exp_timeout;
	    return PyExpect_Err();
	}
	
	/*
	** try to match patterns
	*/
	if (numready > 0) {
	    for (index = 0; index < numproccases; index++) {
		/*
		** there is input for this file descriptor, try to match its
		** patterns
		*/
		if (FD_ISSET(procCases[index].ep->exp_fd, &spawnReadyFdSet)) {
		    exp_match_max    = procCases[index].ep->exp_match_max;
		    exp_loguser      = procCases[index].ep->exp_loguser;
		    exp_full_buffer  = procCases[index].ep->exp_full_buffer;
		    exp_remove_nulls = procCases[index].ep->exp_remove_nulls;
		    result = exp_expectv(procCases[index].ep->exp_fd,
					 procCases[index].ecases);
		    
		    /* Update the buffer and match strings */
		    Py_XDECREF(procCases[index].ep->exp_buffer);
		    if (exp_buffer_end == exp_buffer)
			/* The buffer may be empty, after a
	    		   timeout for example */
			procCases[index].ep->exp_buffer = NULL;
		    else
			procCases[index].ep->exp_buffer =
			    PyString_FromStringAndSize(exp_buffer,
					       exp_buffer_end - exp_buffer);
		    
		    if (exp_match_end == exp_buffer)
			/* No match */
			procCases[index].ep->exp_match_index =
			    procCases[index].ep->exp_match_len =
				procCases[index].ep->exp_match_end = 0;
		    else {
			procCases[index].ep->exp_match_index =
			    exp_match - exp_buffer;
			procCases[index].ep->exp_match_len   =
			    exp_match_end - exp_match;
			procCases[index].ep->exp_match_end   =
			    exp_match_end - exp_buffer;
		    }
		    
		    if (result == EXP_EOF) {
			retSpawnObj = (PyObject *) procCases[index].ep;
			freeProcCaseMem(procCases, numproccases);
			exp_timeout = old_exp_timeout;
			PyErr_SetObject(PyExpect_Error_Eof,
					retSpawnObj);
			return NULL;
		    } else if (result == EXP_FULLBUFFER) {
			retSpawnObj = (PyObject *) procCases[index].ep;
			freeProcCaseMem(procCases, numproccases);
			PyErr_SetObject(PyExpect_Error_FullBuffer, 
					retSpawnObj);
			exp_timeout = old_exp_timeout;
			return NULL;
		    } else if (result >= 0 &&
			       result < procCases[index].numcases) {

			/*
			** Save spawn object to be returned
			*/
			retSpawnObj = (PyObject *) procCases[index].ep;

			freeProcCaseMem(procCases, numproccases);
			exp_timeout = old_exp_timeout;

			return Py_BuildValue("OO", 
					     returnObjArray[result],
					     retSpawnObj);
			/*
			** the only remaining possibility is timing out, 
			** in which case we check the next ready
			** file descriptor
			*/
		    } else if (result != EXP_TIMEOUT) {
			char buffer[40];
			sprintf(buffer, "Unrecognized expect return code %d",
				result);
			PyErr_SetString(PyExpect_Error, buffer);
			freeProcCaseMem(procCases, numproccases);
			exp_timeout = old_exp_timeout;
			return NULL;
		    }

		}
	    }
	}
	
	/* adjust remaining time */
	if (timeout != -1) {
	    newTime = time(NULL);
	    if (newTime == -1) {
		freeProcCaseMem(procCases, numproccases);
		exp_timeout = old_exp_timeout;
		return PyExpect_Err();
	    }
	    
	    timeout -= (newTime - curTime);
	    
	    if (timeout < 0)
		timeout = 0;
	    
	    curTime = newTime;
	}
    } while (timeout > 0 || timeout == -1);
    
    /*
    ** we timed out before matching a pattern
    */
    freeProcCaseMem(procCases, numproccases);
    
    exp_timeout = old_exp_timeout;
    
    PyErr_SetNone(PyExpect_Error_Timeout);
    return NULL;
}

/* Set or get the timeout value */

static PyObject *
PyExpect_SetDefTimeout(PyObject *self, PyObject *args)
{
    int newtimeout, oldtimeout;

    oldtimeout = exp_def_timeout;
    if (PyArg_ParseTuple(args, ""))
	;	/* We don't want to change the value */
    else {
	PyErr_Clear();
	if (PyArg_ParseTuple(args, "i", &newtimeout)) {
	    if (newtimeout >= -1)
		exp_def_timeout = newtimeout;
	    else {
		PyErr_SetString(PyExc_ValueError,
				"Timeout value must be >= -1") ;
		return NULL;
	    }
	} else {
	    PyErr_SetString(PyExc_TypeError,
			    "Timeout parameter must be void or int");
	    return NULL;
	}
    }
    /* We always return the old exp_timeout value */
    return PyInt_FromLong(oldtimeout);
}


/* Set or get the default max no. of chars to match */

static PyObject *
PyExpect_SetDefMatchMax(PyObject *self, PyObject *args)
{
    int newmatchmax, oldmatchmax;

    oldmatchmax = exp_def_match_max;
    if (PyArg_ParseTuple(args, ""))
	;	/* We don't want to change the value */
    else {
	PyErr_Clear();
	if (PyArg_ParseTuple(args, "i", &newmatchmax)) {
	    if (newmatchmax >= 1)
		exp_def_match_max = newmatchmax;
	    else {
		PyErr_SetString(PyExc_ValueError,
				"Max. no. of chars to match must be >= 1") ;
		return NULL;
	    }
	} else {
	    PyErr_SetString(PyExc_TypeError,
			    "Max. no. of chars to match parameter must be void or int");
	    return NULL;
	}
    }
    /* We always return the old exp_def_match_max value */
    return PyInt_FromLong(oldmatchmax);
}



/* Set or get the loguser flag value */

static PyObject *
PyExpect_SetDefLoguser(PyObject *self, PyObject *args)
{
    int newlog, oldlog;
    
    oldlog = exp_def_loguser;
    if (PyArg_ParseTuple(args, ""))
	;	/* We don't want to change the value */
    else {
	PyErr_Clear();
	if (PyArg_ParseTuple(args, "i", &newlog)) {
	    if (newlog == 0 || newlog == 1)
		exp_def_loguser = newlog;
	    else {
		PyErr_SetString(PyExc_ValueError,
				"Loguser value must be 0 or 1") ;
		return NULL;
	    }
	} else {
	    PyErr_SetString(PyExc_TypeError,
			    "Loguser parameter must be void or int");
	    return NULL;
	}
    }
    /* We always return the old exp_loguser value */
    return PyInt_FromLong(oldlog);
}


/* Set or get the fullbuffer flag value */

static PyObject *
PyExpect_SetDefFullbuffer(PyObject *self, PyObject *args)
{
    int newfull_buffer, oldfull_buffer;
    
    oldfull_buffer = exp_def_full_buffer;
    if (PyArg_ParseTuple(args, ""))
	;	/* We don't want to change the value */
    else {
	PyErr_Clear();
	if (PyArg_ParseTuple(args, "i", &newfull_buffer)) {
	    if (newfull_buffer == 0 || newfull_buffer == 1)
		exp_def_full_buffer = newfull_buffer;
	    else {
		PyErr_SetString(PyExc_ValueError,
				"Full_buffer value must be 0 or 1") ;
		return NULL;
	    }
	} else {
	    PyErr_SetString(PyExc_TypeError,
			    "Full_buffer parameter must be void or int");
	    return NULL;
	}
    }
    /* We always return the old exp_full_buffer value */
    return PyInt_FromLong(oldfull_buffer);
}


/* Set or get the remove_nulls flag value */

static PyObject *
PyExpect_SetDefRemoveNulls(PyObject *self, PyObject *args)
{
    int newremove_nulls, oldremove_nulls;
    
    oldremove_nulls = exp_def_remove_nulls;
    if (PyArg_ParseTuple(args, ""))
	;	/* We don't want to change the value */
    else {
	PyErr_Clear();
	if (PyArg_ParseTuple(args, "i", &newremove_nulls)) {
	    if (newremove_nulls == 0 || newremove_nulls == 1)
		exp_def_remove_nulls = newremove_nulls;
	    else {
		PyErr_SetString(PyExc_ValueError,
				"Remove_nulls value must be 0 or 1") ;
		return NULL;
	    }
	} else {
	    PyErr_SetString(PyExc_TypeError,
			    "Remove_nulls parameter must be void or int");
	    return NULL;
	}
    }
    /* We always return the old exp_remove_nulls value */
    return PyInt_FromLong(oldremove_nulls);
}


/* List of functions exported by this module. */

static PyMethodDef PyExpect_methods[] = {
	{ "spawn",		PyExpect_Spawn,             1,
	  "Spawn a new process under the control of Expect" },
	{ "expect",		PyExpect_Expect,            1,
	  "Match a set of patterns to a set of spawned processes" },
	{ "setdeftimeout",	PyExpect_SetDefTimeout,     1,
	  "Set the default timeout in seconds" },
	{ "setdefmatch_max",	PyExpect_SetDefMatchMax,    1,
	  "Set the default for maximum no. of chars to match" },
	{ "setdefloguser",	PyExpect_SetDefLoguser,     1,
	  "Toggle the default for display of the messages on the screen" },
	{ "setdeffull_buffer",	PyExpect_SetDefFullbuffer,  1,
	  "Toggle the default for use of the full_buffer exception" },
	{ "setdefremove_nulls",	PyExpect_SetDefRemoveNulls, 1,
	  "Toggle the default for removal of nulls from the stream" },
	{ NULL,			NULL }		 /* Sentinel */
};


/* Initialize this module.
   This is called when the first 'import expect' is done,
   via a table in config.c */

void
initexpect()
{
    PyObject *m, *d, *x;
    static char module_doc[] =
	"An implementation of the Expect library";
    
    m = Py_InitModule4("expect", PyExpect_methods, module_doc,
		       (PyObject *)NULL, PYTHON_API_VERSION);
    d = PyModule_GetDict(m);
    
    /* Create the error strings */
    PyExpect_Error = PyString_FromString("expect.error");
    if (PyExpect_Error == NULL || 
	PyDict_SetItemString(d, "error", PyExpect_Error) != 0)
	Py_FatalError("can't define expect.error");
    PyExpect_Error_Timeout = PyString_FromString("expect.timeout");
    if (PyExpect_Error_Timeout == NULL || 
	PyDict_SetItemString(d, "timeout", PyExpect_Error_Timeout) != 0)
	Py_FatalError("can't define expect.timeout");
    PyExpect_Error_Eof = PyString_FromString("expect.eof");
    if (PyExpect_Error_Eof == NULL || 
	PyDict_SetItemString(d, "eof", PyExpect_Error_Eof) != 0)
	Py_FatalError("can't define expect.eof");
    PyExpect_Error_FullBuffer = PyString_FromString("expect.full_buffer");
    if (PyExpect_Error_FullBuffer == NULL || 
	PyDict_SetItemString(d, "full_buffer",
			     PyExpect_Error_FullBuffer) != 0)
	Py_FatalError("can't define expect.full_buffer");
    
    /* Create the constants */
    x = PyInt_FromLong((long)exp_exact);
    PyDict_SetItemString(d, "exact", x);
    x = PyInt_FromLong((long)exp_glob);
    PyDict_SetItemString(d, "glob", x);
    x = PyInt_FromLong((long)exp_regexp);
    PyDict_SetItemString(d, "regexp", x);
    x = PyInt_FromLong((long)exp_null);
    PyDict_SetItemString(d, "null", x);
    
    /* Save the values of the default flags */
    exp_def_timeout      = exp_timeout;
    exp_def_loguser      = exp_loguser;
    exp_def_full_buffer  = exp_full_buffer;
    exp_def_remove_nulls = exp_remove_nulls;
    exp_def_match_max    = exp_match_max;
}
